(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* 1. remove ConstraintPat, ConstraintExp and ConstraintSig *)
(* 2. SimpleDec(d1,i1) SimpleDec(d2,i2) --> SimpleDec(SeqDec([d1,d2],i),i) *)
(*    It's okay without this reduction but for performance. *)
(*    Without this reduction, lots of operations on contexts *)

signature POSTINFER =
sig
val postinfer: Absyn.topdec -> Absyn.topdec
end (* POSTINFER *)

structure PostInfer : POSTINFER =
struct

open Location
open Absyn
open List
open Utils

fun pat p =
  case p
    of AppPat (clid, p', t, i) => AppPat (clid, pat p', t, i)
     | RecordPat (labps, t, i) =>
       RecordPat (map (fn (l,p',i) => (l, pat p', i)) labps, t, i)
     | SubRecordPat (labps, t, i) =>
       SubRecordPat (map (fn (l,p',i) => (l, pat p', i)) labps, t, i)
     | RefPat (p', t, i) => RefPat (pat p', t, i)
     | AsPat (v, p', t, i) => AsPat (v, pat p', t, i)
     | TuplePat (ps, t, i) => TuplePat (map pat ps, t, i)
     | ListPat (ps, t, i) => ListPat (map pat ps, t, i)
     | ArrayPat (ps, t, i) => ArrayPat (map pat ps, t, i)
     | OrPat (ps, t, i) => OrPat (map pat ps, t, i)
     | ConstraintPat (p', _, _) => pat p'
     | _ => p

and exp e =
  case e
    of AppExp (e1, e2, t, i) => AppExp (exp e1, exp e2, t, i)
     | RecordExp (labes, t, i) =>
       RecordExp (map (fn (l, e', i) => (l, exp e', i)) labes, t, i)
     | RecordFieldExp (e', l, t, i) => RecordFieldExp (exp e', l, t, i)
     | ArrayFieldExp (e1, e2, t, i) => ArrayFieldExp (exp e1, exp e2, t, i)
     | UpdateArrayExp (e1, e2, e3, t, i) =>
       UpdateArrayExp (exp e1, exp e2, exp e3, t, i)
     | SubstRecordExp (e1, l, e2, t, i) =>
       SubstRecordExp (exp e1, l, exp e2, t, i)
     | TupleExp (es, t, i) => TupleExp (map exp es, t, i)
     | ListExp (es, t, i) => ListExp (map exp es, t, i)
     | ArrayExp (es, t, i) => ArrayExp (map exp es, t, i)
     | LetExp (d, e', t, i) => LetExp (dec d, exp e', t, i)
     | HandleExp (e', rules, t, i) => HandleExp (exp e', map rule rules, t, i)
     | RaiseExp (e', t, i) => RaiseExp (exp e', t, i)
     | FnExp (fnrules, t, i) =>
       FnExp (map (fn (ps, e', i) => (map pat ps, exp e', i)) fnrules, t, i)
     | AssignExp (e1, e2, t, i) => AssignExp (exp e1, exp e2, t, i)
     | RefExp (e', t, i) => RefExp (exp e', t, i)
     | DeRefExp (e', t, i) => DeRefExp (exp e', t, i)
     | SeqExp (es, t, i) => SeqExp (map exp es, t, i)
     | CaseExp (e', rules, t, i) => CaseExp (exp e', map rule rules, t, i)
     | IfExp (e1, e2, e3, t, i) => IfExp (exp e1, exp e2, exp e3, t, i)
     | WhileExp (e1, e2, t, i) => WhileExp (exp e1, exp e2, t, i)
     | ForExp (vid, e1, e2, e3, e4, t, i) =>
       ForExp (vid, exp e1, exp e2, exp e3, exp e4, t, i)
     | ConstraintExp (e', _, _) => exp e'
     | _ => e

and rule (p, e, i) = (pat p, exp e, i)

and valbind (p, e, i) = (pat p, exp e, i)

and dec d =
  case d
    of ValDec (tyvs, valbs, vel, i) =>
       ValDec (tyvs, map valbind valbs, vel, i)
     | RecValDec (tyvs, valbs, vel, i) =>
       RecValDec (tyvs, map valbind valbs, vel, i)
     | LocalDec (d1, d2, i) => LocalDec (dec d1, dec d2, i)
     | SeqDec (ds, i) => SeqDec (map dec ds, i)
     | _ => d

fun strexp stre =
  case stre
    of VarStr _ => stre
     | StrStr (strd, i) => StrStr (strdec strd, i)
     | SigStr (stre', sige, i) => SigStr (strexp stre', sige, i)
     | FctAppStr (fctid, stres, i) => FctAppStr (fctid, map strexp stres, i)

and strbind (strid, sigeopt, stre, i) = (strid, sigeopt, strexp stre, i)

and strdec strd =
  case strd
    of SimpleDec (d, i) => SimpleDec (dec d, i)
     | StrDec (strbs, i) => StrDec (map strbind strbs, i)
     | SeqStrDec ([SimpleDec(OpenDec _,_), sd as SeqStrDec _], _) => strdec sd
     | SeqStrDec (strds as (SimpleDec _)::_, i) =>
       let (* split': (dec * info) list -> dec list * info *)
           fun split' [] = raise Fail "strange SeqStrDec"
             | split' (res as (_,i1)::_) =
               let val (ds', i2) = foldl (fn (ds,_) (d,i) => (ds@[d],i))
                                         ([],none) res
                   val i' = {loc_start = i1.loc_start, loc_end = i2.loc_end,
                             loc_ghost = false}
               in  (ds',i')
               end
           (* split: strdec list -> (dec * info) list -> strdec list *)
           fun split [] [] = []
             | split [] [(d,i)] = [SimpleDec (d,i)]
             | split [] res =
               let val (ds',i') = split' res
               in  [SimpleDec (SeqDec(ds',i'), i')]
               end
             | split ((SimpleDec (d,i))::sds) [] = split sds [(d,i)]
             | split ((SimpleDec (d,i))::sds) res = split sds (res@[(d,i)])
             | split ((sd as StrDec _)::sds) [] = [sd]@(split sds [])
             | split ((sd as StrDec _)::sds) res =
               let val (ds', i') = split' res
               in  [SimpleDec (SeqDec(ds',i'), i')]@[sd]@(split sds [])
               end
             | split ((SeqStrDec _)::_) _ = raise Fail "strange SeqStrDec"
       in  case split strds []
             of [SimpleDec (d,i)] => SimpleDec (dec d, i)
              | strds' => SeqStrDec (map strdec strds', i)
       end
     | SeqStrDec (strds, i) => SeqStrDec (map strdec strds, i)

fun spec spc =
  case spc
    of IncludeSpec (sige, i) => IncludeSpec (sigexp sige, i)
     | StrSpec (strdscs, i) =>
       StrSpec (map (fn (s,ge,i) => (s,sigexp ge,i)) strdscs, i)
     | SeqSpec (spcs, i) => SeqSpec (map spec spcs, i)
     | _ => spc

and sigexp sige =
  case sige
    of VarSig (sigid, i) => VarSig (sigid, i)
     | SigSig (spc, i) => SigSig (spec spc, i)
     | ConstraintSig (sige', _, i) => sigexp sige'

fun sigbind (sigid, sige, i) = (sigid, sigexp sige, i)

fun sigdec (sigbs, i) = (map sigbind sigbs, i)

fun fctdec (fctid, fctargs, stre, i) =
  (fctid, fctargs, strexp stre, i)

fun topdec td =
  case td
    of Sig (sigd, i) => Sig (sigdec sigd, i)
     | Fct (fctd, i) => Fct (fctdec fctd, i)
     | Str (strd, i) => Str (strdec strd, i)
     | SeqTopDec (topds as (Str(SimpleDec _,_))::_, i) =>
       let (* split': (dec * info) list -> dec list * info *)
           fun split' [] = raise Fail "strange SeqTopDec"
             | split' (res as (_,i1)::_) =
               let val (ds', i2) = foldl (fn (ds,_) (d,i) => (ds@[d],i))
                                         ([],none) res
                   val i' = {loc_start = i1.loc_start, loc_end = i2.loc_end,
                             loc_ghost = false}
               in  (ds',i')
               end
           (* split: topdec list -> (dec * info) list -> topdec list *)
           fun split [] [] = []
             | split [] [(d,i)] = [Str(SimpleDec (d,i),i)]
             | split [] res =
               let val (ds',i') = split' res
               in  [Str(SimpleDec (SeqDec(ds',i'), i'),i')]
               end
             | split ((Str(SimpleDec (d,i),_))::tds) [] = split tds [(d,i)]
             | split ((Str(SimpleDec (d,i),_))::tds) res =
               split tds (res@[(d,i)])
             | split ((SeqTopDec _)::_) _ = raise Fail "strange SeqTopDec"
             | split (td::tds) [] = [td]@(split tds [])
             | split (td::tds) res =
               let val (ds', i') = split' res
               in  [Str(SimpleDec (SeqDec(ds',i'),i'),i')]@[td]@(split tds [])
               end
       in  case split topds []
             of [Str(SimpleDec (d,i),i')] => Str(SimpleDec (dec d, i), i')
              | topds' => SeqTopDec (map topdec topds', i)
       end
     | SeqTopDec (topds, i) => SeqTopDec (map topdec topds, i)

fun postinfer absyn = topdec absyn

end (* PostInfer *)
