(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* simplify val bindings *)
(* val k a = k e         --> val a = e *)
(* val {a,b}={a=e1,b=e2} --> val a = e1 and b = e2 *)
(* val ref a = ref e     --> val a = e  *)
(* val (a,b) = (e1,e2)   --> val a = e1 and b = e2 *)
(* val [a,b] = [e1,e2]   --> val a = e1 and b = e2 *)
(* val [|a,b|]=[|e1,e2|] --> val a = e1 and b = e2 *)
(* val p = (e1;e2;e3)    --> val _ = (e1;e2) and p = e3 *)

signature PREINFER =
sig
val preinfer: String_ast.Ast.topdec -> String_ast.Ast.topdec
end (* PREINFER *)

structure PreInfer : PREINFER =
struct

open String_ast.Ast
open List

exception StrangeAst

(* sameSid: strid -> strid -> bool *)
fun sameSid (sid1,_) (sid2,_) = sid1=sid2
(* sameCon: conlongid -> conlongid -> bool *)
fun sameCon (sids1, (name1,_), _) (sids2, (name2,_), _) =
  (fold_left2 (fn f x y => (sameSid x y) andalso f) true sids1 sids2)
  andalso name1=name2
(* lab2pat: Absyn.labpat list -> Absyn.pat list *)
fun lab2pat labpats =
  map (fn (_,p,_) => p)
      (Sort.list (fn ((x,_),_,_) ((y,_),_,_) => x < y) labpats)
(* lab2exp: Absyn.labexp list -> Absyn.exp list *)
fun lab2exp labexps =
  map (fn (_,e,_) => e)
      (Sort.list (fn ((x,_),_,_) ((y,_),_,_) => x < y) labexps)
(* recPE: labpat list -> labexp list -> pat list * exp list *)
fun recPE labpats labexps = (lab2pat labpats, lab2exp labexps)

fun exp e =
  case e
    of AppExp (e1, e2, i) => AppExp (exp e1, exp e2, i)
     | RecordExp (labes, i) =>
       RecordExp (map (fn (l, e', i) => (l, exp e', i)) labes, i)
     | RecordFieldExp (e', l, i) => RecordFieldExp (exp e', l, i)
     | ArrayFieldExp (e1, e2, i) => ArrayFieldExp (exp e1, exp e2, i)
     | UpdateArrayExp (e1, e2, e3, i) =>
       UpdateArrayExp (exp e1, exp e2, exp e3, i)
     | SubstRecordExp (e1, l, e2, i) => SubstRecordExp (exp e1, l, exp e2, i)
     | TupleExp (es, i) => TupleExp (map exp es, i)
     | ListExp (es, i) => ListExp (map exp es, i)
     | ArrayExp (es, i) => ArrayExp (map exp es, i)
     | LetExp (d, e', i) => LetExp (dec d, exp e', i)
     | HandleExp (e', rules, i) => HandleExp (exp e', map rule rules, i)
     | RaiseExp (e', i) => RaiseExp (exp e', i)
     | FnExp (fnrules, i) =>
       FnExp (map (fn (ps, e', i) => (ps, exp e', i)) fnrules, i)
     | AssignExp (e1, e2, i) => AssignExp (exp e1, exp e2, i)
     | RefExp (e', i) => RefExp (exp e', i)
     | DeRefExp (e', i) => DeRefExp (exp e', i)
     | SeqExp (es, i) => SeqExp (map exp es, i)
     | CaseExp (e', rules, i) => CaseExp (exp e', map rule rules, i)
     | IfExp (e1, e2, e3, i) => IfExp (exp e1, exp e2, exp e3, i)
     | WhileExp (e1, e2, i) => WhileExp (exp e1, exp e2, i)
     | ForExp (vid, e1, e2, e3, e4, i) =>
       ForExp (vid, exp e1, exp e2, exp e3, exp e4, i)
     | ConstraintExp (e', t, i) => ConstraintExp (exp e', t, i)
     | _ => e

and rule (p, e, i) = (p, exp e, i)

and vbs res (ps, es) i = res@(map2 (fn p e => (p, e, i)) ps es)

(* val k a = k e         --> val a = e *)
and valbind res (AppPat (c,p,_), AppExp (ConExp(c',_),e,_), i) =
    if sameCon c c' then res@[(p, exp e, i)] else raise StrangeAst
(* val {a,b}={a=e1,b=e2} --> val a = e1 and b = e2 *)
  | valbind res (RecordPat (labp,_), RecordExp (labe,_), i) =
    vbs res (recPE labp labe) i
(* val ref a = ref e     --> val a = e  *)
  | valbind res (RefPat (p,_), RefExp (e,_), i) = res@[(p, exp e, i)]
(* val (a,b) = (e1,e2)   --> val a = e1 and b = e2 *)
  | valbind res (TuplePat (ps,_), TupleExp (es,_), i) = vbs res (ps, es) i
(* val [a,b] = [e1,e2]   --> val a = e1 and b = e2 *)
  | valbind res (ListPat (ps,_), ListExp (es,_), i) = vbs res (ps, es) i
(* val [|a,b|]=[|e1,e2|] --> val a = e1 and b = e2 *)
  | valbind res (ArrayPat (ps,_), ArrayExp (es,_), i) = vbs res (ps, es) i
(* val p = (e1;e2;e3)    --> val _ = (e1;e2) and p = e3 *)
  | valbind res (p, SeqExp ([],_), i) = raise Failure "strange SeqExp"
  | valbind res (p, SeqExp ([e],_), i) = valbind res (p, e, i)
  | valbind res (p, SeqExp (es,_), i) =
    let fun lstLast [] res = raise Failure "strange SeqExp"
          | lstLast [x] res = (res, x)
          | lstLast (x::xs) res = lstLast xs (res@[x])
        val (front, last) = lstLast es nil
    in  res@[(WildPat i, SeqExp (map exp front,i), i), (p, exp last, i)]
    end
  | valbind res (p, e, i) = res@[(p, exp e, i)]

and funbind (bds, i) = (map body bds, i)

and body (vid, ps, e, i) = (vid, ps, exp e, i)

and dec d =
  case d
    of ValDec (tyvs, valbs, i) => ValDec (tyvs, fold_left valbind [] valbs, i)
     | RecValDec (tyvs, valbs, i) =>
       RecValDec (tyvs, fold_left valbind [] valbs, i)
     | FunDec (tyvs, funbs, i) => FunDec (tyvs, map funbind funbs, i)
     | AbstypeDec (tybs, d', i) => AbstypeDec (tybs, dec d', i)
     | LocalDec (d1, d2, i) => LocalDec (dec d1, dec d2, i)
     | SeqDec (ds, i) => SeqDec (map dec ds, i)
     | _ => d

fun strexp stre =
  case stre
    of VarStr _ => stre
     | StrStr (strd, i) => StrStr (strdec strd, i)
     | SigStr (stre', sige, i) => SigStr (strexp stre', sige, i)
     | FctAppStr (fctid, stres, i) => FctAppStr (fctid, map strexp stres, i)

and strbind (strid, sigeopt, stre, i) = (strid, sigeopt, strexp stre, i)

and strdec strd =
  case strd
    of SimpleDec (d, i) => SimpleDec (dec d, i)
     | StrDec (strbs, i) => StrDec (map strbind strbs, i)
     | SeqStrDec (strds, i) => SeqStrDec (map strdec strds, i)

fun fctdec (fctid, fctargs, sigeopt, stre, i) =
  (fctid, fctargs, sigeopt, strexp stre, i)

fun topdec td =
  case td
    of Sig (sigd, i) => Sig (sigd, i)
     | Fct (fctd, i) => Fct (fctdec fctd, i)
     | Str (strd, i) => Str (strdec strd, i)
     | SeqTopDec (topds, i) => SeqTopDec (map topdec topds, i)

fun preinfer ast = topdec ast

end (* PreInfer *)
