(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature S =
sig
type elt
type t
val empty: t
val is_empty: t -> bool
val mem: elt -> t -> bool
val add: elt -> t -> t
val singleton: elt -> t
val union: t -> t -> t
val inter: t -> t -> t
val diff: t -> t -> t
val subset: t -> t -> bool
val iter: (elt -> unit) -> t -> unit
val fold: (elt -> 'a -> 'a) -> t -> 'a -> 'a
val cardinal: t -> int
val elements: t -> elt list
end (* S *)

signature M =
sig
type key
type 'a t
val empty: 'a t
val add: key -> 'a -> 'a t -> 'a t
val find: key -> 'a t -> 'a
val remove: key -> 'a t -> 'a t
val iter: (key -> 'a -> unit) -> 'a t -> unit
val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
end (* M *)

signature H =
sig
type key
type 'a t
val create: int -> 'a t
val add: 'a t -> key -> 'a -> unit
val find: 'a t -> key -> 'a
val iter: (key -> 'a -> unit) -> 'a t -> unit
end (* H *)

signature UTILS =
sig
structure IS : S where type elt = int
structure SS : S where type elt = string
structure LS : S where type elt = Location.t
structure ES : S where type elt = string * Location.t
structure IM : M where type key = int
structure SM : M where type key = string
structure TM : M where type key = Ty.ty
structure LM : M where type key = Location.t
structure EM : M where type key = string * Location.t
structure SH : H where type key = string
structure T : sig val time: string -> (unit -> 'a) -> 'a
                  val pp: unit -> unit
              end
val list2IS: int list -> IS.t
val inc: int ref -> unit
exception Fail of string
val ($$): ('a -> 'b) -> ('c -> 'a) -> 'c -> 'b
val foldl: ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
val print: string -> unit
val i2s: int -> string
val b2s: bool -> string
val r2s: real -> string
val take: ('a list * int) -> 'a list
val drop: ('a list * int) -> 'a list
val lstMinus: 'a list -> 'a list -> 'a list
val lstlast: 'a list -> 'a list -> 'a list * 'a

val cfa: bool ref
val debugflag: bool ref
val debug: string -> unit

end (* UTILS *)

structure Utils : UTILS =
struct

structure L = List

structure Int = struct type t = int val compare = compare end
structure Str = struct type t = string val compare = compare end
structure Exn = struct type t = string * Location.t val compare = compare end
structure Fty = struct type t = Ty.ty val compare = compare end
structure Loc = struct type t = Location.t val compare = compare end

structure IS = MakeSet(Int)
structure SS = MakeSet(Str)
structure LS = MakeSet(Loc)
structure ES = MakeSet(Exn)

structure IM = MakeMap(Int)
structure SM = MakeMap(Str)
structure TM = MakeMap(Fty)
structure LM = MakeMap(Loc)
structure EM = MakeMap(Exn)

structure HStr =
struct type t = string val equal = (=) val hash = Hashtbl.hash end
structure SH = MakeHash(HStr)

(* list2IS: int list -> IS.t *)
fun list2IS [] = IS.empty
  | list2IS (x::xs) = IS.union (IS.singleton x) (list2IS xs)

(* inc: int ref -> unit *)
fun inc cnt = cnt := !cnt + 1

structure T =
struct
  (* m: (real ref * int ref) SH.t *)
  (* m: function name |-> (time, count) *)
  val m = SH.create 10
  (* time: string -> (unit -> 'a) -> 'a *)
  fun time s f =
    let val start = Unix.gettimeofday()
	val r = f ()
	val endt = Unix.gettimeofday()
        val (tm, cnt) = (SH.find m s)
	                handle Not_found => let val c = (ref 0.0, ref 0)
				            in  SH.add m s c; c end
    in  (tm := (((endt:real) - start):real) + (!tm); inc cnt; r)
    end
  open Format
  (* pp: unit -> unit *)
  fun pp () =
    let val tms = ref ([]: (string * int * real) list)
    in  (SH.iter (fn s (tm,cnt) => tms := (s,!cnt,!tm)::(!tms)) m;
	 tms := L.sort (fn (s1,_,_) (s2,_,_) => compare s1 s2) (!tms);
	 L.iter (fn (s,c,t) =>
		    printf "\n[%10s: %d times, %.3f seconds]" s c t) (!tms);
	 printf "@.")
    end
end

exception Fail of string

(* $$ : ('a -> 'b) -> ('c -> 'a) -> 'c -> 'b *)
fun ($$) f g = fn x => f (g x)

(* foldl: ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a *)
val foldl = L.fold_left

(* print: string -> unit *)
val print = print_string

(* i2s: int -> string *)
val i2s = string_of_int
(* b2s: bool -> string *)
val b2s = string_of_bool
(* r2s: real -> string *)
val r2s = string_of_float

(* take: ('a list * int) -> 'a list *)
fun take (lst, n) =
  let fun loop (l, 0) res = res
        | loop ([], _) _ = raise Fail "Subscript"
	| loop ((h::t), n) res = loop (t, n-1) (res@[h])
  in  if n >= 0 then loop (lst, n) nil else raise Fail "Subscript"
  end

(* drop: ('a list * int) -> 'a list *)
fun drop (lst, n) =
  let fun loop (l, 0) = l
        | loop ([], _) = raise Fail "Subscript"
        | loop ((_::t), n) = loop (t, n-1)
  in  if n >= 0 then loop (lst, n) else raise Fail "Subscript"
  end

(* lstMinus: 'a list -> 'a list -> 'a list *)
(* lstMinus [3,6,2,4,6,9] [4,6,9];; *)
(* - : int list = [3,6,2] *)
fun lstMinus lst1 lst2 = take (lst1, (L.length lst1)-(L.length lst2))

(* lstlast: 'a list -> 'a list -> 'a list * 'a *)
fun lstlast [] res = raise Fail "last: empty list"
  | lstlast [x] res = (res, x)
  | lstlast (x::xs) res = lstlast xs (res@[x])

(* cfa: bool ref *)
val cfa = ref false
(* debugflag: bool ref *)
val debugflag = ref false
(* debug: string -> unit *)
fun debug msg = if !debugflag then print ("\n"^msg)

end (* Utils *)
